var TabExpose = {
  bExposed: false,
  NumTabs: 1,
  NumColumns: 1,

  DoExpose: function() {
    if(!this.bExposed) {
      // do expose
      this.NumTabs = getBrowser().tabContainer.childNodes.length;
      if(this.NumTabs == 1) // no poin
        return;

      this.CalculatePreviewSizes();
      this.CreateAllCanvasPopups();
      this.ShowAllPopups();
      this.bExposed = true;
    }
    else {
      this.CancelExpose();
    }
  },

  CancelExpose: function() {
    window.removeEventListener("click", this.CancelExpose, false);
    this.HideAllPopups();
    this.bExposed = false;
  },

  WindowClickListener: function() {
    var TP = this;
    return function() {
      TP.CancelExpose();
    };
  },

  // Show all the popups in a nice grid.
  //TODO: animate their appearance.
  // maybe fade in, maybe zoom to position,
  // maybe both?
  ShowAllPopups: function() {
    var ps = document.getElementById("tabexposePopups");
    var x,y,r,c;
    for(var i=0; i<this.NumTabs; i++) {
      r = Math.floor(i / this.NumColumns);
      c = i % this.NumColumns;
      x = /*window.screenX + */ (2 * c + 1) * this.HorizontalSpacing/2 + c*this.PreviewSize;
      y = /*window.screenY + */ (2 * r + 1) * this.VerticalSpacing/2 + r*this.PreviewSize;
      //ps.childNodes[i].showPopup(window.document.firstChild, x, y, "popup")
      ps.childNodes[i].style.top = Math.round(y) + "px";
      ps.childNodes[i].style.left = Math.round(x) + "px";
      ps.childNodes[i].style.display = "block";
    }

    // this is in case you click elsewhere
    window.addEventListener("click", this.WindowClickListener(), false);
  },

  // hide all of our popup previews
  HideAllPopups: function() {
    var ps = document.getElementById("tabexposePopups");
    for(var i=this.NumTabs-1; i>=0; i--) {
      ps.childNodes[i].style.display = "none";
    }
  },

  // figure out how many tabs we have,
  // how many rows and columns we ought to organize them into,
  // and what size each should be.
  CalculatePreviewSizes: function() {
    var s = 1;
    for(var x=2; s<this.NumTabs; s=x*x, x++)
      ;

    this.NumColumns = x;
    // now figure out sizes.  we want square previews!
    var size = Math.min(window.innerWidth / this.NumColumns,
                        window.innerHeight / this.NumColumns);
    // shrink it a little bit to get some spacing
    size = Math.floor(size * 0.9);
    this.PreviewSize = size;
    this.VerticalSpacing = (window.innerHeight - (size * this.NumColumns)) / this.NumColumns;
    this.HorizontalSpacing = (window.innerWidth - (size * this.NumColumns)) / this.NumColumns;
  },
  
  // Make sure we have enough popups to show all our tabs.
  // If not, create extras.
  // Update the size of each popup and draw its tab contents.
  CreateAllCanvasPopups: function () {
    var b = getBrowser();
    var ps = document.getElementById("tabexposePopups");

    // first add extra tabs if we need them
    if(this.NumTabs > ps.childNodes.length) {
      for(var i=ps.childNodes.length; i<this.NumTabs; i++) {
        ps.appendChild(this.CreateCanvasPopup());
      }
    }

    // now update sizes and draw contents
    for(var i=0; i<this.NumTabs; i++) {
      var c = ps.childNodes[i].firstChild;
      ps.childNodes[i].setAttribute("tooltip", b.getBrowserAtIndex(i).contentWindow.title);
      c.style.width = this.PreviewSize + "px";
      c.style.height = this.PreviewSize + "px";
      c.width = c.height = this.PreviewSize;
      var cx = c.getContext("2d");
      this.DrawPreview(b.getBrowserAtIndex(i).contentWindow, cx);
      
      // add click handling
      c.onclick = this.CreateClickHandler(b, b.tabContainer.childNodes[i]);
    }
  },

  // create a click handler function that sets this tabbrowser's
  // current tab to tab
  CreateClickHandler: function(browser, tab) {
    var TP = this;
    return function() {
      dump(tab.linkedBrowser.contentWindow.location.href + "\n");
      browser.selectedTab = tab;
      TP.CancelExpose();
    };
  },

  // create a popup containing a html:canvas
  CreateCanvasPopup: function() {
    var box = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul", "box");
    //TODO: put this in a css file somewhere?
    box.style.border = "1px solid grey";
    box.style.position = "fixed";
    box.style.top = box.style.left = 0;
    box.style.display = "none";
    var c = document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");
    box.appendChild(c);
    return box;
  },

  // draw a preview of the contents of win
  // into the 2d context cx
  DrawPreview: function(win, cx) {
    var w = win.innerWidth;
    var h = win.innerHeight;
    
    // set the preview initially for top left
    var xScroll = 0;
    var yScroll = 0;
    var fudge = 0;
    
    // if user has set preference for scrolled preview, get the scrolled position
    if (1) {
      xScroll = win.scrollX;
      yScroll = win.scrollY;
    }
    else {
      // bug 313178
      fudge = win.scrollMaxY;
    }
    
    try {
      //var d = new Date().getTime();
      cx.clearRect(0, 0, this.PreviewSize, this.PreviewSize);
      cx.save();
      cx.scale(this.PreviewSize/w, this.PreviewSize/h);
      cx.drawWindow(win, xScroll, yScroll, w, h+fudge, "rgb(255,255,255)");
      cx.restore();
      //d = (new Date().getTime()) - d;
      //dump("Took " + d + " ms\n");
    }
    catch(e) { dump('TabExpose Error: ' + e.message + '\n'); }
  }
};
